//-------------------------------------------------------------------
// File: cbproj/MsgTester/MsgTestU1.cpp
// Purpose: First "Unit" with a "form" (window) of a 'tester'
//          for DL4YHF's inter-application  communication module.
//       Uses windows' WM_COPYDATA-Message mechanism
//       to pass data from one application to another.
//       More info in ..\YHF_Tools\YHF_Comm.cpp, where most of the
//       "protocol" is implemented.
//
//       (c) 2001..2004 by Wolfgang Buescher
//
// Author: Wolfgang Buescher, (www.qsl.net/DL4YHF)
//
// Revision history (YYYY-MM-DD):
//  2020-10-29 :
//        Tried to add a COMMAND LINE INTERFACE for this tool.
//        Also tried to link this application to create an executable
//        *without* any external files (like DLLs); more on that below.
// 
//  2004-10-01 :
//        Minor corrections (YHF_FindWindowByName) + clarifications
//  2001-08-27 :
//        Started as private experiment in DL4YHF's Spectrum Lab.
//
//---------------------------------------------------------------------------
//
//  How to compile / link this application into an executable WITHOUT
//   *any* external files (DLLs etc etc) :
//   From http://bcbjournal.org/articles/vol4/0009/Building_stand-alone_EXEs.htm :
// > Creating a stand-alone EXE
// > Creating a stand-alone EXE requires modifying the project options to use
// > static linking. There are two project options you must modify to accomplish this.
// > Those options are explained in the following sections.
// > After you have changed the project options and rebuild the project,
// > the EXE will not require external files in order to execute.
// >
// > The dynamic RTL
// > The first setting you must modify is the projects use of the C++ runtime
// > library DLL. This DLL contains the code for the C++ library.
// > All C++ applications need the RTL code in order to operate.
// > By default, a new C++Builder project is configured to use the RTL DLL.
// > The RTL DLL for C++Builder 5 is named CC3250MT.DLL. For other versions
// > of C++Builder, it has a slightly different name. For C++Builder 4,
// > for example, the DLL is named CP3245MT.DLL.
// >
// > To eliminate the need for the runtime library DLL, open the
// > Project Options dialog, click on the Linker tab,
// > and uncheck the Use dynamic RTL option.
// >
// >
// > Runtime packages
// >
// > The second setting you must modify is the setting that controls
// > how packages are used by the application. The default project settings
// > require the use of runtime packages.
// >
// > All VCL components (including those written by third parties) must reside
// > in a package. Packages are nearly identical to DLLs in the way they are
// > used by an application (the technical differences are not pertinent here).
// > When you build an application using runtime packages, the application
// > must be able to load the runtime package when it starts. When the use
// > of runtime packages is turned off, the code for all components
// > used in the project is linked into the executable.
// > To turn off the use of runtime packages, select the Packages page
// > on the Project Options dialog and uncheck (!!!) the
// > "Build with runtime packages" option.
//
// DL4YHF, 2020-10 : Been there, done that, et voila:
//         The size of the executable (MsgTest1.exe) grew from
//     32 to  88 kByte after unchecking "use dynamic RTL",   and from
//     88 to 542 kByte (!) after also unchecking "build with runtime packages".
//     Quite a price to pay for such a microscopic application, but anyway...         
#include <vcl.h>
#pragma hdrstop

#include "YHF_Comm.h"     // DL4YHF's Inter-Application Messages
#include "YHF_CmdLine.h"  // Utilities to retrieve arguments from the command line

#include "MsgTestU1.h"

// Use standard "C", not Borland VCL types wherever you can :
BOOL fBatchMode  = FALSE; // flag: FALSE= not in batch mode (run until USER closes us),
                          //       TRUE = running in batch mode, closing in a few milliseconds.
int  iBatchTimer = 0;     // number of 100-millisecond-intervals ...
int  iBatchTimeout = 10;  // .. for command-line-driven "batch mode" (since 2020)


//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMsgTestForm *MsgTestForm;
//---------------------------------------------------------------------------
__fastcall TMsgTestForm::TMsgTestForm(TComponent* Owner)
        : TForm(Owner)
{
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
void __fastcall TMsgTestForm::Btn_SendMessageClick(TObject *Sender)
{
 DWORD         dw,hDestWindow;
 LRESULT       lSendResult;
 T_YHF_COMMAND yhfCommandMsg;

 // ex: (replaced with YHF_FindWindowByName)
 // hDestWindow = (DWORD)FindWindow(
 //   Ed_SendMsgName->Text.c_str(), // LPCTSTR lpClassName = pointer to class name
 //   NULL); // LPCTSTR lpWindowName = pointer to window name (title, not unique)
 hDestWindow = YHF_FindWindowByName( Ed_SendMsgName->Text.c_str() );
 if(hDestWindow==0)
   {
     Memo_Messages->Lines->Add(
      "Cannot find a window named \""
       +Ed_SendMsgName->Text+"\" !");
     return;
   }

  yhfCommandMsg.dwSender     = (DWORD)Handle; // MY WINDOW HANDLE
  yhfCommandMsg.cu.c.Module1 = Ed_ModuleID->Text.c_str()[0];
  yhfCommandMsg.cu.c.Module2 = Ed_ModuleID->Text.c_str()[1];
  yhfCommandMsg.cu.c.Command1 = Ed_Command->Text.c_str()[0];
  yhfCommandMsg.cu.c.Command2 = Ed_Command->Text.c_str()[1];
  strcpy(yhfCommandMsg.du.cData, Ed_MsgText->Text.c_str());
  yhfCommandMsg.lDataSize = // Laenge des Blocks incl. Nullbyte
      strlen(yhfCommandMsg.du.cData)+1;

  lSendResult = YHF_SendCommandMessage(
            hDestWindow,  // handle of destination window
        &yhfCommandMsg ); // pointer to message being sent

  Lab_SendMsgResult->Caption = "-> " + IntToStr( (__int64)lSendResult);
  //  Memo_Messages->Lines->Add(
  //   "Result from SendMessage = "+IntToStr( (__int64)lSendResult) );
}
//---------------------------------------------------------------------------




void __fastcall TMsgTestForm::FormCreate(TObject *Sender)
{
  int  i, nCommandLineArguments, iLen;
  char szCmdArg[256], *cp;

  m_CopydataMessageCounter = 0;

  // Botschafts-Behandlungs-Routine auf niedrigem Level umbiegen,
  //  um Windozw-Messages wie WM_COPYDATA abfangen zu koennen .
  // Hinweis: die schrottige Methode TApplication.OnMessage kann nur
  //          Messages empfangen, die gePOSTET wurden, aber nicht geSENDet !
  WindowProc = MyWindowProc;


  // Parse the command line. Maybe someone invokes us from a BATCH FILE,
  //  let us send something to someone, and then return immediately ?
  nCommandLineArguments = YHF_GetCmdLineParamCount();
  // Note: ParamStr(0) is the name of the executable, which is useless here ...
  i = 1; // ... thus start with the ZERO-BASED index for the SECOND(!) argument
  while( i<=/*!*/nCommandLineArguments )
   { // ex: strcpy(cmd_string, ParamStr(i).c_str() );  // removed, because...
     // Caution, Borland's ParamStr(i) completely sucks ! ! !
     //  For example, it turned  /sp.print("hello")
     //               into       /sp.print(hello)
     //   which is something completely different (hello is a variable in the 2nd case)
     // The reason seems to be a bug in Borland's VCL ("ParamStr").
     YHF_GetCmdLineParamStr( i, szCmdArg, sizeof(szCmdArg)-1 );
     YHF_RemoveOuterDoubleQuotes( szCmdArg );    // ..with emphasis on "Outer".. !
     iLen = strlen( szCmdArg );
     if( szCmdArg[1]=='=' && (iLen>2) )    // looks like a key=value pair, e.g. m=x u=x t=x :
      {
        cp = szCmdArg+2;
        switch(szCmdArg[0]) // branch by the single-character "key" ..
         { case 'n' :  // n=<Name of the recipient>, e.g. n=SpecLab1, SpecLab2, etc.
              // Copy the rest of this argument into the edit field labelled
              //    'Window name for SendMessage()' :
              Ed_SendMsgName->Text = AnsiString(szCmdArg+2); // copy everything after the '='
              break;
           case 'm' :  // m=<Module Name>, almost always "CL" for "Commad Line [Interpreter]"
              // Copy the rest of this argument into the edit field labelled
              //    'Module ID' :
              Ed_ModuleID->Text = AnsiString(szCmdArg+2);
              break;
           case 'c' :  // c=<Command Type>, e.g. "CF"=Calculate Function or "EC"=Execute Command.
              // Copy the rest of this argument into the edit field labelled "Command".
              // When not specified, 'batch mode' will use the default ("CF"),
              // and in that case, the batch-mode-timeout should be long enough
              // to SEE THE RESPONSE in the GUI. Unfortunately, there's no easy
              // way to return the response text from a "CF"-command through
              // the crippled operation system (the value returned by an application,
              // like the value returned from the good old main()-function in C,
              // is an INTEGER, not a STRING).
              Ed_Command->Text = AnsiString(szCmdArg+2);
              fBatchMode = TRUE;  // definitely running in timer-controlled "batch mode" now
              if( Ed_Command->Text == "CF" ) // "Call Function" ?
               { // In this particular case, we expect another WM_COPYDATA message
                 // with the results. So let the 'batch mode' wait a bit longer,
                 // or until the response has arrived.
                 // ALL OTHER "command types" (e.g. "EC"="Execute Command")
                 // are not expected to "return a result", so for THEM,
                 // batch mode will time-out after the default (10 * 100 ms) .
                 iBatchTimeout = 30/* timer-ticks a 100 ms*/;
               }
              break;

           // ToDo: Implement your own 'keys' here .
           default:
              break;
         } // end switch(cmd_string[0])
      } // end < looks like a PARAMETER >
     else if(szCmdArg[0]=='\\' || szCmdArg[0]=='/') // slash or backslash : "string to send"
      {
         // Copy the rest of this argument into the edit field labelled "Text":
         Ed_MsgText->Text = AnsiString(szCmdArg+1); // copy everything after the slash (or backslash)
         fBatchMode = TRUE;  // definitely running in timer-controlled "batch mode" now
      } // end if  < \xxx or /xxx in command line >
     else  // anything else (not a single-character "key", or string beginning with backslash) :
      { cp = strrchr(szCmdArg, '.');   // maybe a filename, look for the extension:
        if(cp)
         { // ToDo : Add your own processing of certain file types, e.g.
           // if( stricmp(cp,".txt")==0)
           //  { // looks like a TEXT FILE so send the entire content ...
           //  }
         } // end if < found an extension >
      } // end < argument doesn't look like a "parameter" >
     ++i; // next argument please...
   } // end  while( i<=/*!*/nCommandLineArguments )

  if( fBatchMode )
   { Memo_Messages->Lines->Add( "Running in batch mode, closing in "
           + IntToStr( iBatchTimeout ) + " ms ." );
   }
} // end TMsgTestForm::FormCreate()
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
void __fastcall TMsgTestForm::MyWindowProc(Messages::TMessage &Message)
  // low level message handler.
  // Installed in the OnCreate method of the application's main window.
  // Receives both SENT and POSTED window messages
  //   (in contrast to that lousy TApplication.OnMessage thing...)
{
  // The WM_COPYDATA message is sent when an application passes data
  //                 to another application.
  //
  // WM_COPYDATA
  // wParam = (WPARAM) (HWND) hwnd;            // handle of sending window
  // lParam = (LPARAM) (PCOPYDATASTRUCT) pcds; // pointer to structure with data
  //
  // Parameters
  //   hwnd   Identifies the window passing the data.
  //   pcds   Points to a COPYDATASTRUCT structure that contains
  //          the data to be passed.
  // Return Values
  //   If the receiving application processes this message,
  //   it should return TRUE; otherwise, it should return FALSE.
  //
  // Remarks
  //   An application must use the SendMessage function to send this message,
  //   not the PostMessage function.
  //   The data being passed must not contain pointers or other references
  //   to objects not accessible to the application receiving the data.
  //   While this message is being sent, the referenced data must not
  //   be changed by another thread of the sending process.
  //   The receiving application should consider the data read-only.
  //   The pcds parameter is valid only during the processing of the message.
  //   The receiving application should not free the memory referenced by pcds.
  // ! If the receiving application must access the data after SendMessage
  // ! returns, it must copy the data into a local buffer.
 if (Message.Msg == WM_COPYDATA)
  {
    ++m_CopydataMessageCounter;
    Message.Result  = FALSE;  // return value if message couldn't be handled
    if(Message.LParam != NULL)
     { // only if the pointer to a COPYDATASTRUCT struct exists:
       // typedef struct tagCOPYDATASTRUCT {  // cds
       //    DWORD dwData;    // Specifies up to 32 bits of data
       //                     // to be passed to the receiving application.
       //    DWORD cbData;    // Specifies the size, in bytes, of the data
       //                     // pointed to by the lpData member.
       //    PVOID lpData;    // Points to data to be passed to the receiving
       //                     // application. This member can be NULL.
       //  } COPYDATASTRUCT;
       Message.Result = YHF_HandleCopydataMessage(
        Message.WParam, // =  hwnd, identifies the window passing the data.
        ((PCOPYDATASTRUCT)Message.LParam)->dwData,  // 32 bits of data (eg command)
        ((PCOPYDATASTRUCT)Message.LParam)->cbData,  // size of datablock in bytes
        ((PCOPYDATASTRUCT)Message.LParam)->lpData); // pointer to datablock

     }
  } // end if Message.Msg == WM_COPYDATA
 else
  WndProc(Message); // WndProc ist die "Original-Prozedur" fuers Message-Handling
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
void __fastcall TMsgTestForm::HandleCommandMessages(void)
   // Reads messages from external applications from an internal queue,
   // and executes all commands for registered modules.
{
 T_YHF_COMMAND *pCmd;
 char str255[256];
 char *cp;

  while( (pCmd=YHF_PeekCommandMessage( '*','*'  // MODULE id: wildcards (ALL)
                        ,FALSE ) ) /*don't remove yet*/
            != NULL ) // message received ?
   {
     cp= str255;
     strcpy(cp,"id=\""); cp+=strlen(cp);
     if(pCmd->cu.c.Module1)
        *cp++ = pCmd->cu.c.Module1;
     if(pCmd->cu.c.Module2)
        *cp++ = pCmd->cu.c.Module2;
     *cp++=',';
     if(pCmd->cu.c.Command1)
        *cp++ = pCmd->cu.c.Command1;
     if(pCmd->cu.c.Command2)
        *cp++ = pCmd->cu.c.Command2;
     strcpy(cp,"\",dat="); cp+=strlen(cp);
     if( (pCmd->lDataSize>0) && (pCmd->lDataSize<200) )
       { strncpy(cp, (char*)pCmd->du.bData, pCmd->lDataSize);
         cp[pCmd->lDataSize] = 0;
       }
     Memo_Messages->Lines->Add(str255);

     YHF_RemoveCommandMessage(pCmd);

     if( fBatchMode )
      { iBatchTimeout = 0;  // no need to wait any longer in TMsgTestForm::Timer1Timer()
      }
   } // end while ( ..PeekCommandMessage.. )
} // end CTEST_HandleCommandMessages
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
void __fastcall TMsgTestForm::Timer1Timer(TObject *Sender)
  // Periodically invoked from somewhere (Borland's VCL) every 100 milliseconds
  // (timer interval specified in the 'Object Inspector' as property 'Interval').
{
  Lab_CountRcvd->Caption = IntToStr(m_CopydataMessageCounter);

  HandleCommandMessages();

  if( fBatchMode )  // still running in "batch mode" ?
   { ++iBatchTimer; // another 100 milliseconds have passed ...

     switch( iBatchTimer )  // in the N-th call of this timer-event-handler...
      { case 1 :    // .. in the FIRST call, imiate clicking "Send Message" :
           Btn_SendMessageClick( Sender );
           break;

        default:    // Nothing else to do, just wait a little longer for the RESPONSE.
           break;

      } // end switch( iBatchTimer )

     if( iBatchTimer > iBatchTimeout ) // batch mode finished (timed out) ?
      { Close(); // <- a Borland VCL thing, closes the main window and exits
      }
   }

} // end TMsgTestForm::Timer1Timer()
//---------------------------------------------------------------------------


